import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import dill
import base64
import httpx
from typing import Any, List, Dict, Optional, Type, Union
from pydantic import BaseModel
import os
import tempfile

from io import BytesIO


class Others:
    def screenshot(self, show: bool = True, save_path: Optional[str] = None) -> Optional[bytes]:
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg
        """
        Take a screenshot using the server and optionally display it or save it.

        Args:
            show: Whether to display the screenshot using matplotlib
            save_path: Optional path to save the screenshot

        Returns:
            The screenshot bytes if save_path is not provided
        """
        # Get the screenshot from the server
        response = self.send_request("/others/take_screenshot", {}, method="GET", return_raw=True)
        
        if save_path:
            # Save the screenshot to the specified path
            with open(save_path, 'wb') as f:
                f.write(response)
        
        if show:
            # Display the screenshot using matplotlib
            img = mpimg.imread(BytesIO(response))
            plt.figure(figsize=(15, 10))
            plt.axis('off')
            plt.imshow(img)
            plt.show()
        
        if not save_path:
            return response
        
        return None



import os
import sqlite3
import json
from dotenv import load_dotenv
import signal
from .folder import BASE_PATH


class ConfigManager:
    def __init__(self, db_name="config.sqlite"):
        self.db_path = os.path.join(BASE_PATH, db_name)
        self.conn = sqlite3.connect(self.db_path)
        self.cursor = self.conn.cursor()
        
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS config_store (
                key TEXT PRIMARY KEY,
                value TEXT NOT NULL
            )
        ''')
        self.conn.commit()
        
        signal.signal(signal.SIGTERM, self._handle_signal)
        signal.signal(signal.SIGINT, self._handle_signal)

    def _handle_signal(self, signum, frame):
        if hasattr(self, 'conn'):
            try:
                self.conn.commit()
                self.conn.close()
            except:
                pass
        signal.default_int_handler(signum, frame)

    def initialize(self, key):
        load_dotenv()
        value = os.getenv(key)
        if value is not None:
            self.set(key, value)

    def get(self, key, default=None):
        try:
            self.cursor.execute('SELECT value FROM config_store WHERE key = ?', (key,))
            result = self.cursor.fetchone()
            return json.loads(result[0]) if result else default
        except:
            return default

    def delete(self, key):
        try:
            self.cursor.execute('DELETE FROM config_store WHERE key = ?', (key,))
            self.conn.commit()
            return self.cursor.rowcount > 0
        except:
            return False

    def set(self, key, value):
        try:
            value_json = json.dumps(value)
            self.cursor.execute('REPLACE INTO config_store (key, value) VALUES (?, ?)',
                              (key, value_json))
            self.conn.commit()
            return True
        except:
            return False

    def dump(self):
        try:
            self.conn.commit()
            return True
        except:
            return False

    def __del__(self):
        if hasattr(self, 'conn'):
            try:
                self.conn.commit()
                self.conn.close()
            except:
                pass


# Create a single instance of ConfigManager
Configuration = ConfigManager()

Configuration.initialize("OPENAI_API_KEY")
Configuration.initialize("ANTHROPIC_API_KEY")
Configuration.initialize("AZURE_OPENAI_ENDPOINT")
Configuration.initialize("AZURE_OPENAI_API_VERSION")
Configuration.initialize("AZURE_OPENAI_API_KEY")
Configuration.initialize("AWS_ACCESS_KEY_ID")
Configuration.initialize("AWS_SECRET_ACCESS_KEY")
Configuration.initialize("AWS_REGION")
Configuration.initialize("DEEPSEEK_API_KEY")

ClientConfiguration = ConfigManager(db_name="client_config.sqlite")


import asyncio
import os
from typing import ClassVar, Literal

from anthropic.types.beta import BetaToolBash20241022Param

from .base import BaseAnthropicTool, CLIResult, ToolError, ToolResult


class _BashSession:
    """A session of a bash shell."""

    _started: bool
    _process: asyncio.subprocess.Process

    command: str = "/bin/bash"
    _output_delay: float = 0.2  # seconds
    _timeout: float = 120.0  # seconds
    _sentinel: str = "<<exit>>"

    def __init__(self):
        self._started = False
        self._timed_out = False

    async def start(self):
        if self._started:
            return

        self._process = await asyncio.create_subprocess_shell(
            self.command,
            preexec_fn=os.setsid,
            shell=True,
            bufsize=0,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        self._started = True

    def stop(self):
        """Terminate the bash shell."""
        if not self._started:
            raise ToolError("Session has not started.")
        if self._process.returncode is not None:
            return
        self._process.terminate()

    async def run(self, command: str):
        """Execute a command in the bash shell."""
        if not self._started:
            raise ToolError("Session has not started.")
        if self._process.returncode is not None:
            return ToolResult(
                system="tool must be restarted",
                error=f"bash has exited with returncode {self._process.returncode}",
            )
        if self._timed_out:
            raise ToolError(
                f"timed out: bash has not returned in {self._timeout} seconds and must be restarted",
            )

        # we know these are not None because we created the process with PIPEs
        assert self._process.stdin
        assert self._process.stdout
        assert self._process.stderr

        # send command to the process
        self._process.stdin.write(
            command.encode() + f"; echo '{self._sentinel}'\n".encode()
        )
        await self._process.stdin.drain()

        # read output from the process, until the sentinel is found
        try:
            async with asyncio.timeout(self._timeout):
                while True:
                    await asyncio.sleep(self._output_delay)
                    # if we read directly from stdout/stderr, it will wait forever for
                    # EOF. use the StreamReader buffer directly instead.
                    output = self._process.stdout._buffer.decode()  # pyright: ignore[reportAttributeAccessIssue]
                    if self._sentinel in output:
                        # strip the sentinel and break
                        output = output[: output.index(self._sentinel)]
                        break
        except asyncio.TimeoutError:
            self._timed_out = True
            raise ToolError(
                f"timed out: bash has not returned in {self._timeout} seconds and must be restarted",
            ) from None

        if output.endswith("\n"):
            output = output[:-1]

        error = self._process.stderr._buffer.decode()  # pyright: ignore[reportAttributeAccessIssue]
        if error.endswith("\n"):
            error = error[:-1]

        # clear the buffers so that the next output can be read correctly
        self._process.stdout._buffer.clear()  # pyright: ignore[reportAttributeAccessIssue]
        self._process.stderr._buffer.clear()  # pyright: ignore[reportAttributeAccessIssue]

        return CLIResult(output=output, error=error)


class BashTool(BaseAnthropicTool):
    """
    A tool that allows the agent to run bash commands.
    The tool parameters are defined by Anthropic and are not editable.
    """

    _session: _BashSession | None
    name: ClassVar[Literal["bash"]] = "bash"
    api_type: ClassVar[Literal["bash_20241022"]] = "bash_20241022"

    def __init__(self):
        self._session = None
        super().__init__()

    async def __call__(
        self, command: str | None = None, restart: bool = False, **kwargs
    ):
        if restart:
            if self._session:
                self._session.stop()
            self._session = _BashSession()
            await self._session.start()

            return ToolResult(system="tool has been restarted.")

        if self._session is None:
            self._session = _BashSession()
            await self._session.start()

        if command is not None:
            return await self._session.run(command)

        raise ToolError("no command provided.")

    def to_params(self) -> BetaToolBash20241022Param:
        return {
            "type": self.api_type,
            "name": self.name,
        }


from fastapi import HTTPException
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Union
import traceback
from ...api import app, timeout

import asyncio
from concurrent.futures import ThreadPoolExecutor
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import base64
from ....storage.configuration import Configuration


prefix = "/storage"



class ConfigGetRequest(BaseModel):
    key: str

class ConfigSetRequest(BaseModel):
    key: str
    value: str

class BulkConfigSetRequest(BaseModel):
    configs: Dict[str, str]


@app.post(f"{prefix}/config/get")
async def get_config(request: ConfigGetRequest):
    """
    Endpoint to get a configuration value by key using POST.

    Args:
        key: The configuration key

    Returns:
        The configuration value or a default message if not found
    """
    value = Configuration.get(request.key)
    return {"key": request.key, "value": value}


@app.post(f"{prefix}/config/set")
async def set_config(request: ConfigSetRequest):
    """
    Endpoint to set a configuration value.

    Args:
        key: The configuration key
        value: The configuration value

    Returns:
        A success message
    """
    Configuration.set(request.key, request.value)
    return {"message": "Configuration updated successfully"}

@app.post(f"{prefix}/config/bulk_set")
async def bulk_set_config(request: BulkConfigSetRequest):
    """
    Endpoint to set multiple configuration values at once.

    Args:
        configs: Dictionary of configuration key-value pairs

    Returns:
        A success message
    """
    for key, value in request.configs.items():
        Configuration.set(key, value)
    return {"message": "Bulk configuration updated successfully"}




from typing import Any
from decimal import Decimal
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.align import Align
from .price import get_estimated_cost
import platform



console = Console()

# Global dictionary to store aggregated values by price_id
price_id_summary = {}

def spacing():
    console.print("")



def connected_to_server(server_type: str, status: str, total_time: float = None):
    """
    Prints a 'Connected to Server' section for Upsonic, full width,
    with two columns: 
      - left column (labels) left-aligned
      - right column (values) left-aligned, positioned on the right half 
    """

    # Determine color and symbol for the status
    if status.lower() == "established":
        status_text = "[green]✓ Established[/green]"
    elif status.lower() == "failed":
        status_text = "[red]✗ Failed[/red]"
    else:
        status_text = f"[cyan]… {status}[/cyan]"

    # Build a table that expands to full console width
    table = Table(show_header=False, expand=True, box=None)
    
    # We define 2 columns, each with ratio=1 so they evenly split the width
    # Both are left-aligned, but the second column will end up on the right half.
    table.add_column("Label", justify="left", ratio=1)
    table.add_column("Value", justify="left", ratio=1)

    # Rows: one for server type, one for status
    table.add_row("[bold]Server Type:[/bold]", f"[yellow]{server_type}[/yellow]")
    table.add_row("[bold]Connection Status:[/bold]", status_text)
    
    if total_time is not None:
        table.add_row("[bold]Total Time:[/bold]", f"[cyan]{total_time:.2f} seconds[/cyan]")

    table.width = 60

    # Wrap the table in a Panel that also expands full width
    panel = Panel(
        table, 
        title="[bold cyan]Upsonic - Server Connection[/bold cyan]",
        border_style="cyan",
        expand=True,  # panel takes the full terminal width
        width=70  # Adjust as preferred
    )

    # Print the panel (it will fill the entire width, with two columns inside)
    console.print(panel)

    spacing()

def call_end(result: Any, llm_model: str, response_format: str, start_time: float, end_time: float, usage: dict, debug: bool = False):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]LLM Model:[/bold]", f"{llm_model}")
    # Add spacing
    table.add_row("")

    from ..client.level_two.agent import SubTaskList, SearchResult, CompanyObjective, HumanObjective

    is_it_subtask = isinstance(result, SubTaskList)
    is_it_search = isinstance(result, SearchResult)
    is_it_company = isinstance(result, CompanyObjective)
    is_it_human = isinstance(result, HumanObjective)

    if is_it_subtask:
        # Print total task count
        table.add_row(f"[bold]Total Subtasks:[/bold]", f"[yellow]{len(result.sub_tasks)}[/yellow]")
        table.add_row("")
        # Print each task as well as bullet list
        for each in result.sub_tasks:
            table.add_row(f"[bold]Subtask:[/bold]", f"[green]{each.description}[/green]")
            table.add_row(f"[bold]Required Output:[/bold]", f"[green]{each.required_output}[/green]")
            table.add_row(f"[bold]Tools:[/bold]", f"[green]{each.tools}[/green]")
            table.add_row("")
    elif is_it_search:
        table.add_row("[bold]Has Customers:[/bold]", f"[green]{'Yes' if result.any_customers else 'No'}[/green]")
        table.add_row("")
        table.add_row("[bold]Products:[/bold]")
        for product in result.products:
            table.add_row("", f"[green]• {product}[/green]")
        table.add_row("")
        table.add_row("[bold]Services:[/bold]")
        for service in result.services:
            table.add_row("", f"[green]• {service}[/green]")
        table.add_row("")
        table.add_row("[bold]Potential Competitors:[/bold]")
        for competitor in result.potential_competitors:
            table.add_row("", f"[yellow]• {competitor}[/yellow]")
        table.add_row("")
    elif is_it_company:
        table.add_row("[bold]Company Objective:[/bold]", f"[blue]{result.objective}[/blue]")
        table.add_row("")
        table.add_row("[bold]Goals:[/bold]")
        for goal in result.goals:
            table.add_row("", f"[blue]• {goal}[/blue]")
        table.add_row("")
        table.add_row("[bold]State:[/bold]", f"[blue]{result.state}[/blue]")
        table.add_row("")
    elif is_it_human:
        table.add_row("[bold]Job Title:[/bold]", f"[magenta]{result.job_title}[/magenta]")
        table.add_row("")
        table.add_row("[bold]Job Description:[/bold]", f"[magenta]{result.job_description}[/magenta]")
        table.add_row("")
        table.add_row("[bold]Job Goals:[/bold]")
        for goal in result.job_goals:
            table.add_row("", f"[magenta]• {goal}[/magenta]")
        table.add_row("")
    else:
        result_str = str(result)
        # Limit result to 370 characters
        if not debug:
            result_str = result_str[:370]
        # Add ellipsis if result is truncated
        if len(result_str) < len(str(result)):
            result_str += "[bold white]...[/bold white]"

        table.add_row("[bold]Result:[/bold]", f"[green]{result_str}[/green]")

    # Add spacing
    table.add_row("")
    table.add_row("[bold]Response Format:[/bold]", f"{response_format}")

    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)}$")
    time_taken = end_time - start_time
    time_taken_str = f"{time_taken:.2f} seconds"
    table.add_row("[bold]Time Taken:[/bold]", f"{time_taken_str}")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Call Result[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()



def agent_end(result: Any, llm_model: str, response_format: str, start_time: float, end_time: float, usage: dict, tool_count: int, context_count: int, debug: bool = False, price_id:str = None):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    # Track values if price_id is provided
    if price_id:
        estimated_cost = get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)
        if price_id not in price_id_summary:
            price_id_summary[price_id] = {
                'input_tokens': 0,
                'output_tokens': 0,
                'estimated_cost': 0.0
            }
        price_id_summary[price_id]['input_tokens'] += usage['input_tokens']
        price_id_summary[price_id]['output_tokens'] += usage['output_tokens']
        price_id_summary[price_id]['estimated_cost'] = Decimal(str(price_id_summary[price_id]['estimated_cost'])) + Decimal(str(estimated_cost).replace('$', '').replace('~', ''))

    table.add_row("[bold]LLM Model:[/bold]", f"{llm_model}")
    # Add spacing
    table.add_row("")
    result_str = str(result)
    # Limit result to 370 characters
    if not debug:
        result_str = result_str[:370]
    # Add ellipsis if result is truncated
    if len(result_str) < len(str(result)):
        result_str += "[bold white]...[/bold white]"

    table.add_row("[bold]Result:[/bold]", f"[green]{result_str}[/green]")
    # Add spacing
    table.add_row("")
    table.add_row("[bold]Response Format:[/bold]", f"{response_format}")
    
    table.add_row("[bold]Tools:[/bold]", f"{tool_count} [bold]Context Used:[/bold]", f"{context_count}")
    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)}$")
    time_taken = end_time - start_time
    time_taken_str = f"{time_taken:.2f} seconds"
    table.add_row("[bold]Time Taken:[/bold]", f"{time_taken_str}")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Agent Result[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()


def agent_total_cost(total_input_tokens: int, total_output_tokens: int, total_time: float, llm_model: str):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(total_input_tokens, total_output_tokens, llm_model)}$")
    table.add_row("[bold]Time Taken:[/bold]", f"{total_time:.2f} seconds")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Agent Total Cost[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )
    console.print(panel)
    spacing()

def print_price_id_summary(price_id: str, task) -> dict:
    """
    Get the summary of usage and costs for a specific price ID and print it in a formatted panel.
    
    Args:
        price_id (str): The price ID to look up
        
    Returns:
        dict: A dictionary containing the usage summary, or None if price_id not found
    """
    if price_id not in price_id_summary:
        console.print("[bold red]Price ID not found![/bold red]")
        return None
    
    summary = price_id_summary[price_id].copy()
    # Format the estimated cost to include $ symbol
    summary['estimated_cost'] = f"${summary['estimated_cost']:.4f}"

    # Create a table for pretty printing
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Price ID:[/bold]", f"[magenta]{price_id}[/magenta]")
    table.add_row("")  # Add spacing
    table.add_row("[bold]Input Tokens:[/bold]", f"[magenta]{summary['input_tokens']:,}[/magenta]")
    table.add_row("[bold]Output Tokens:[/bold]", f"[magenta]{summary['output_tokens']:,}[/magenta]")
    table.add_row("[bold]Total Estimated Cost:[/bold]", f"[magenta]{summary['estimated_cost']}[/magenta]")

    panel = Panel(
        table,
        title="[bold magenta]Upsonic - Price ID Summary[/bold magenta]",
        border_style="magenta",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()

    return summary

def agent_retry(retry_count: int, max_retries: int):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Retry Status:[/bold]", f"[yellow]Attempt {retry_count + 1} of {max_retries + 1}[/yellow]")
    
    panel = Panel(
        table,
        title="[bold yellow]Upsonic - Agent Retry[/bold yellow]",
        border_style="yellow",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()

def get_price_id_total_cost(price_id: str):
    """
    Get the total cost for a specific price ID.
    
    Args:
        price_id (str): The price ID to get totals for
        
    Returns:
        dict: Dictionary containing input tokens, output tokens, and estimated cost for the price ID.
        None: If the price ID is not found.
    """
    if price_id not in price_id_summary:
        return None

    data = price_id_summary[price_id]
    return {
        'input_tokens': data['input_tokens'],
        'output_tokens': data['output_tokens'],
        'estimated_cost': float(data['estimated_cost'])
    }

def missing_dependencies(tool_name: str, missing_deps: list):
    """
    Prints a formatted panel with missing dependencies and installation instructions.
    
    Args:
        tool_name: Name of the tool with missing dependencies
        missing_deps: List of missing dependency names
    """
    if not missing_deps:
        return
    
    # Create the installation command
    install_cmd = "pip install " + " ".join(missing_deps)
    
    # Create a bulleted list of dependencies
    deps_list = "\n".join([f"  • [bold white]{dep}[/bold white]" for dep in missing_deps])
    
    # Create the content with the dependencies list and installation command
    content = f"[bold red]Missing Dependencies for {tool_name}:[/bold red]\n\n{deps_list}\n\n[bold green]Installation Command:[/bold green]\n  {install_cmd}"
    
    # Create and print the panel
    panel = Panel(content, title="[bold yellow]⚠️ Dependencies Required[/bold yellow]", border_style="yellow", expand=False)
    console.print(panel)

def missing_api_key(tool_name: str, env_var_name: str, dotenv_support: bool = True):
    """
    Prints a formatted panel with information about a missing API key and how to set it.
    
    Args:
        tool_name: Name of the tool requiring the API key
        env_var_name: Name of the environment variable for the API key
        dotenv_support: Whether the tool supports loading from .env file
    """
    # Determine the operating system
    system = platform.system()
    
    # Create OS-specific instructions for setting the API key
    if system == "Windows":
        env_instructions = f"setx {env_var_name} your_api_key_here"
        env_instructions_temp = f"set {env_var_name}=your_api_key_here"
        env_description = f"[bold green]Option 1: Set environment variable (Windows):[/bold green]\n  • Permanent (new sessions): {env_instructions}\n  • Current session only: {env_instructions_temp}"
    else:  # macOS or Linux
        env_instructions_export = f"export {env_var_name}=your_api_key_here"
        env_instructions_profile = f"echo 'export {env_var_name}=your_api_key_here' >> ~/.bashrc  # or ~/.zshrc"
        env_description = f"[bold green]Option 1: Set environment variable (macOS/Linux):[/bold green]\n  • Current session: {env_instructions_export}\n  • Permanent: {env_instructions_profile}"
    
    if dotenv_support:
        dotenv_instructions = f"Create a .env file in your project directory with:\n  {env_var_name}=your_api_key_here"
        content = f"[bold red]Missing API Key for {tool_name}[/bold red]\n\n[bold white]The {env_var_name} environment variable is not set.[/bold white]\n\n{env_description}\n\n[bold green]Option 2: Use a .env file:[/bold green]\n  {dotenv_instructions}"
    else:
        content = f"[bold red]Missing API Key for {tool_name}[/bold red]\n\n[bold white]The {env_var_name} environment variable is not set.[/bold white]\n\n{env_description}"
    
    # Create and print the panel
    panel = Panel(content, title="[bold yellow]🔑 API Key Required[/bold yellow]", border_style="yellow", expand=False)
    console.print(panel)


from pydantic import BaseModel
from typing import Dict, Any, Any
import httpx
import time
import asyncio
import concurrent.futures
import threading


from .level_one.call import Call
from .level_two.agent import Agent
from .tasks.tasks import Task
from .agent_configuration.agent_configuration import AgentConfiguration
from .storage.storage import Storage, ClientConfig
from .tools.tools import Tools
from .markdown.markdown import Markdown
from .others.others import Others
from ..exception import ServerStatusException, TimeoutException

from .printing import connected_to_server


from .latest_upsonic_client import latest_upsonic_client


# Helper function to run a coroutine in a new thread with a new event loop
def run_coroutine_in_new_thread(coro):
    """
    Run a coroutine in a new thread with a new event loop.
    This is useful when we're in an async context but need a synchronous result.
    
    Args:
        coro: The coroutine to run
        
    Returns:
        The result of the coroutine
    """
    def run_coro_in_thread(coro):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(coro)
        finally:
            loop.close()
    
    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        return executor.submit(run_coro_in_thread, coro).result()


# Create a base class with url
class UpsonicClient(Call, Storage, Tools, Agent, Markdown, Others):

    def __init__(self, url: str, debug: bool = False, **kwargs):
        """Initialize the Upsonic client.
        
        Args:
            url: The server URL to connect to
            debug: Whether to enable debug mode
            **kwargs: Configuration options that match ClientConfig fields
        """
        start_time = time.time()
        self.debug = debug

        # Set server type and URL first
        if "0.0.0.0" in url:
            self.server_type = "Local(Docker)"
        elif "localhost" in url:
            self.server_type = "Local(Docker)"
        elif "upsonic.ai" in url:
            self.server_type = "Cloud(Upsonic)"
        elif "devserver" in url or "localserver" in url:
            self.server_type = "Local(LocalServer)"
        else:
            self.server_type = "Cloud(Unknown)"

        # Handle local server setup
        if url == "devserver" or url == "localserver":
            url = "http://localhost:7541"
            from ..server import run_dev_server, stop_dev_server, is_tools_server_running, is_main_server_running
            if debug:
                run_dev_server(redirect_output=False)
            else:
                run_dev_server(redirect_output=True)

            import atexit
            def exit_handler():
                if is_tools_server_running() or is_main_server_running():
                    stop_dev_server()
            atexit.register(exit_handler)

        # Set URL and default model
        self.url = url
        self.default_llm_model = "openai/gpt-4o"

        # Check if we're in an async context
        try:
            loop = asyncio.get_running_loop()
            in_async_context = True
        except RuntimeError:
            in_async_context = False

        # Check server status before proceeding
        if in_async_context:
            # We're in an async context, but __init__ can't be async
            # We need to run the async method in a new thread
            status_ok = run_coroutine_in_new_thread(self.status_async())
        else:
            # We're not in an async context, use asyncio.run
            status_ok = asyncio.run(self.status_async())
            
        if not status_ok:
            total_time = time.time() - start_time
            connected_to_server(self.server_type, "Failed", total_time)
            raise ServerStatusException("Failed to connect to the server at initialization.")
        
        # Handle configuration through ClientConfig model
        config = ClientConfig(**(kwargs or {}))
        
        # Create a dictionary of non-None values
        config_dict = {
            key: str(value) for key, value in config.model_dump().items() 
            if value is not None
        }
        
        # Bulk set the configurations if there are any
        if config_dict:
            if in_async_context:
                # We're in an async context, but __init__ can't be async
                # We need to run the async method in a new thread
                run_coroutine_in_new_thread(self.bulk_set_config_async(config_dict))
            else:
                # We're not in an async context, use asyncio.run
                asyncio.run(self.bulk_set_config_async(config_dict))

        global latest_upsonic_client
        latest_upsonic_client = self
        total_time = time.time() - start_time
        connected_to_server(self.server_type, "Established", total_time)

    def status(self) -> bool:
        """Check the server status."""
        try:
            # Try to get the current event loop
            try:
                loop = asyncio.get_running_loop()
                if loop.is_running():
                    # We're in an async context, but this is a sync method
                    # We need to run the async method in a new thread
                    return run_coroutine_in_new_thread(self.status_async())
            except RuntimeError:
                # No event loop is running, use asyncio.run
                return asyncio.run(self.status_async())
        except httpx.RequestError:
            return False

    async def status_async(self) -> bool:
        """Check the server status asynchronously."""
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(self.url + "/status")
                return response.status_code == 200
        except httpx.RequestError:
            return False

    def send_request(self, endpoint: str, data: Dict[str, Any], files: Dict[str, Any] = None, method: str = "POST", return_raw: bool = False) -> Any:
        """
        General method to send an API request.

        Args:
            endpoint: The API endpoint to send the request to.
            data: The data to send in the request.
            files: Optional files to upload.
            method: HTTP method to use (GET or POST)
            return_raw: Whether to return the raw response content instead of JSON

        Returns:
            The response from the API, either as JSON or raw content.
        """
        try:
            # Try to get the current event loop
            try:
                loop = asyncio.get_running_loop()
                if loop.is_running():
                    # We're in an async context, but this is a sync method
                    # We need to run the async method in a new thread
                    return run_coroutine_in_new_thread(
                        self.send_request_async(endpoint, data, files, method, return_raw)
                    )
            except RuntimeError:
                # No event loop is running, use asyncio.run
                return asyncio.run(self.send_request_async(endpoint, data, files, method, return_raw))
        except httpx.RequestError as e:
            raise e

    async def send_request_async(self, endpoint: str, data: Dict[str, Any], files: Dict[str, Any] = None, method: str = "POST", return_raw: bool = False) -> Any:
        """
        Asynchronous version of send_request.
        General method to send an API request asynchronously.

        Args:
            endpoint: The API endpoint to send the request to.
            data: The data to send in the request.
            files: Optional files to upload.
            method: HTTP method to use (GET or POST)
            return_raw: Whether to return the raw response content instead of JSON

        Returns:
            The response from the API, either as JSON or raw content.
        """
        async with httpx.AsyncClient() as client:
            if method.upper() == "GET":
                response = await client.get(self.url + endpoint, params=data, timeout=600.0)
            else:
                if files:
                    response = await client.post(self.url + endpoint, data=data, files=files, timeout=600.0)
                else:
                    response = await client.post(self.url + endpoint, json=data, timeout=600.0)
                
            if response.status_code == 408:
                raise TimeoutException("Request timed out")
            response.raise_for_status()
            
            return response.content if return_raw else response.json()

    def run(self, *args, **kwargs):
        """
        Run method that delegates to the appropriate async implementation.
        """
        try:
            # Try to get the current event loop
            try:
                loop = asyncio.get_running_loop()
                if loop.is_running():
                    # We're in an async context, but this is a sync method
                    # We need to run the async method in a new thread
                    return run_coroutine_in_new_thread(
                        self.run_async(*args, **kwargs)
                    )
            except RuntimeError:
                # No event loop is running, use asyncio.run
                return asyncio.run(self.run_async(*args, **kwargs))
        except Exception as e:
            raise e

    async def run_async(self, *args, **kwargs):
        """
        Asynchronous version of the run method.
        """
        llm_model = kwargs.get("llm_model", None)

        # If there is an two positional arguments we will run it in self.agent_async(first argument, second argument)
        if len(args) == 2:
            
            if isinstance(args[0], AgentConfiguration) and isinstance(args[1], Task):
                return await self.agent_async(args[0], args[1])
            elif isinstance(args[0], list):
                return await self.multi_agent_async(args[0], args[1])
        

        if len(args) == 1:
            return await self.call_async(args[0], llm_model=llm_model)


import os
from upsonic.storage.configuration import ConfigManager


def test_initialize(monkeypatch):
    # Mock environment variable
    monkeypatch.setenv("TEST_KEY", "test_value")
    config = ConfigManager(db_name="test_config.sqlite")
    config.initialize("TEST_KEY")
    assert config.get("TEST_KEY") == "test_value"


def test_set_and_get():
    config = ConfigManager(db_name="test_config.sqlite")
    config.set("key1", "value1")
    assert config.get("key1") == "value1"


def test_get_with_default():
    config = ConfigManager(db_name="test_config.sqlite")
    assert config.get("non_existent_key", "default_value") == "default_value"


def teardown_module(module):
    # Clean up the test database file
    if os.path.exists("test_config.sqlite"):
        os.remove("test_config.sqlite")



from ..agent_configuration.agent_configuration import get_or_create_client, register_tools
from ..tasks.tasks import Task
from typing import Any, Callable, TypeVar, cast

T = TypeVar('T')


class DirectStatic:
    """Static methods for making direct LLM calls using the Upsonic client."""
    
    @staticmethod
    def do(task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Execute a direct LLM call with the given task and model.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            
        Returns:
            The response from the LLM
        """
        global latest_upsonic_client
        from ..latest_upsonic_client import latest_upsonic_client

        # Use provided client or get/create one
        if client is not None:
            the_client = client
        else:
            the_client = get_or_create_client(debug=debug)
        
        # Register tools if needed
        the_client = register_tools(the_client, task.tools)

        # Execute the direct call
        return the_client.call(task, model)

    @staticmethod
    def print_do(task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Execute a direct LLM call and print the result.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            
        Returns:
            The response from the LLM
        """
        result = DirectStatic.do(task, model, client, debug)
        print(result)
        return result


class DirectInstance:
    """Instance-based class for making direct LLM calls using the Upsonic client."""
    
    def __init__(self, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Initialize a DirectInstance with specific model and client settings.
        
        Args:
            model: The LLM model to use (default: None)
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
        """
        self.model = model
        self.client = client
        self.debug = debug
    
    def do(self, task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Execute a direct LLM call using instance defaults or overrides.
        
        Args:
            task: The task to execute
            model: The LLM model to use (overrides instance default if provided)
            client: Optional custom client (overrides instance default if provided)
            debug: Whether to enable debug mode (overrides instance default if provided)
            
        Returns:
            The response from the LLM
        """
        # Use provided parameters or instance defaults
        actual_model = model if model is not None else self.model
        actual_client = client if client is not None else self.client
        actual_debug = debug if debug is not False else self.debug
        
        # Call the static method with the resolved parameters
        return DirectStatic.do(task, actual_model, actual_client, actual_debug)
        
    def print_do(self, task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Execute a direct LLM call and print the result.
        
        Args:
            task: The task to execute
            model: The LLM model to use (overrides instance default if provided)
            client: Optional custom client (overrides instance default if provided)
            debug: Whether to enable debug mode (overrides instance default if provided)
            
        Returns:
            The response from the LLM
        """
        result = self.do(task, model, client, debug)
        print(result)
        return result


class Direct:
    """
    Router class that provides both static and instance-based approaches for direct LLM calls.
    
    When used without initialization, it provides static methods.
    When initialized with parameters, it returns an instance-based object.
    """
    
    # Static methods that delegate to DirectStatic
    @staticmethod
    def do(task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        return DirectStatic.do(task, model, client, debug)
    
    @staticmethod
    def print_do(task: Task, model: str | None = None, client: Any = None, debug: bool = False):
        return DirectStatic.print_do(task, model, client, debug)
    
    def __new__(cls, model: str | None = None, client: Any = None, debug: bool = False):
        """
        Factory method that returns a DirectInstance object when initialized.
        
        Args:
            model: The LLM model to use (default: None)
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            
        Returns:
            A DirectInstance object
        """
        return DirectInstance(model, client, debug)



from fastapi import FastAPI, HTTPException, Request, Response
import asyncio
from functools import wraps
from ...exception import TimeoutException
import inspect
from starlette.responses import JSONResponse
from concurrent.futures import ThreadPoolExecutor
import threading
import time


from fastapi import FastAPI, HTTPException, Request, Response
import asyncio
from functools import wraps
from ...exception import TimeoutException
import inspect
from starlette.responses import JSONResponse
import signal
from concurrent.futures import ThreadPoolExecutor
import threading
import time

app = FastAPI()

# Import the cleanup function from server_utils instead of tools
from .server_utils import cleanup_all_servers

@app.on_event("shutdown")
async def shutdown_event():
    """
    Clean up all server instances when the application shuts down.
    """
    await cleanup_all_servers()


async def timeout_handler(duration: float, coro):
    try:
        return await asyncio.wait_for(coro, timeout=duration)
    except asyncio.TimeoutError:
        raise TimeoutException(f"Operation timed out after {duration} seconds")
def timeout(seconds: float):
    def decorator(func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            def handler(signum, frame):
                raise TimeoutException(f"Function timed out after {seconds} seconds")

            # Set the signal handler and a timeout
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(int(seconds))

            try:
                if inspect.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
                return result
            except TimeoutException as e:
                raise HTTPException(
                    status_code=408,
                    detail=str(e)
                )
            finally:
                # Disable the alarm
                signal.alarm(0)

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            def handler(signum, frame):
                raise TimeoutException(f"Function timed out after {seconds} seconds")

            # Set the signal handler and a timeout
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(int(seconds))

            try:
                return func(*args, **kwargs)
            except TimeoutException as e:
                raise HTTPException(
                    status_code=408,
                    detail=str(e)
                )
            finally:
                # Disable the alarm
                signal.alarm(0)

        return async_wrapper if inspect.iscoroutinefunction(func) else sync_wrapper
    return decorator

@app.get("/status")
async def get_status():
    return {"status": "Server is running"}



"""
Module for handling caching of data using SQLite.
"""

import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import dill
import base64
import time
from typing import Optional, Any
from .configuration import ClientConfiguration


def save_to_cache_with_expiry(data: Any, cache_key: str, expiry_seconds: int) -> None:
    """
    Save data to cache with expiration time.
    
    Args:
        data: Any data to store in cache
        cache_key: Unique identifier for the cached data
        expiry_seconds: Number of seconds until the cache expires
    """
    the_module = dill.detect.getmodule(data)
    if the_module is not None:
        cloudpickle.register_pickle_by_value(the_module)
        
    current_time = int(time.time())
    expiry_time = current_time + expiry_seconds
    cache_key_full = f"cache_{cache_key}"
    
    cache_data = {
        'data': data,
        'expiry_time': expiry_time,
        'created_at': current_time
    }
    
    try:
        ClientConfiguration.delete(cache_key_full)
        serialized_data = base64.b64encode(cloudpickle.dumps(cache_data)).decode('utf-8')
        ClientConfiguration.set(cache_key_full, serialized_data)
    except Exception:
        ClientConfiguration.delete(cache_key_full)
        raise


def get_from_cache_with_expiry(cache_key: str) -> Optional[Any]:
    """
    Retrieve data from cache if not expired.
    
    Args:
        cache_key: Unique identifier for the cached data
        
    Returns:
        Cached data if found and not expired, None otherwise
    """
    cache_key_full = f"cache_{cache_key}"
    serialized_data = ClientConfiguration.get(cache_key_full)

    if serialized_data is None:
        return None
    
    try:
        cache_data = cloudpickle.loads(base64.b64decode(serialized_data))
        current_time = int(time.time())
        
        if current_time > cache_data['expiry_time']:
            ClientConfiguration.delete(cache_key_full)
            return None

        return cache_data['data']
    except Exception:
        ClientConfiguration.delete(cache_key_full)
        return None


from fastapi import HTTPException
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Union
import traceback

import pydantic_ai
from ...api import app, timeout
from ..agent import Agent
import asyncio
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import base64


prefix = "/level_two"


class AgentRequest(BaseModel):
    agent_id: str
    prompt: str
    images: Optional[List[str]] = None
    response_format: Optional[Any] = []
    tools: Optional[Any] = []
    context: Optional[Any] = None
    llm_model: Optional[Any] = "openai/gpt-4o"
    system_prompt: Optional[Any] = None
    retries: Optional[Any] = 1
    context_compress: Optional[Any] = False
    memory: Optional[Any] = False


@app.post(f"{prefix}/agent")
@timeout(500.0)  # 5 minutes timeout for AI operations
async def call_agent(request: AgentRequest):
    """
    Endpoint to call GPT-4 with optional tools and MCP servers.

    Args:
        request: GPT4ORequest containing prompt and optional parameters

    Returns:
        The response from the AI model
    """
    try:
        # Handle pickled response format
        if request.response_format != "str":
            try:
                # Decode and unpickle the response format
                pickled_data = base64.b64decode(request.response_format)
                response_format = cloudpickle.loads(pickled_data)
            except Exception as e:
                # Fallback to basic type mapping if unpickling fails
                type_mapping = {
                    "str": str,
                    "int": int,
                    "float": float,
                    "bool": bool,
                }
                response_format = type_mapping.get(request.response_format, str)
        else:
            response_format = str

        if request.context is not None:
            try:
                pickled_context = base64.b64decode(request.context)
                context = cloudpickle.loads(pickled_context)
            except Exception as e:
                context = None
        else:
            context = None

        result = await Agent.agent(
            agent_id=request.agent_id,
            prompt=request.prompt,
            images=request.images,
            response_format=response_format,
            tools=request.tools,
            context=context,
            llm_model=request.llm_model,
            system_prompt=request.system_prompt,
            retries=request.retries,
            context_compress=request.context_compress,
            memory=request.memory
        )

        if request.response_format != "str" and result["status_code"] == 200:
            result["result"] = cloudpickle.dumps(result["result"])
            result["result"] = base64.b64encode(result["result"]).decode('utf-8')
        return {"result": result, "status_code": 200}

    except pydantic_ai.exceptions.UnexpectedModelBehavior as e:
        return {"result": {"status_code": 500, "detail": f"Change your response format to a simple format or improve your task description. Your response format is too hard for the model to understand (Dont use 'dict' like things in your response format, define everything explicitly). Try to make it more small parts.", "status_code": 500}}

    except Exception as e:
        traceback.print_exc()
        return {"result": {"status_code": 500, "detail": f"Error processing Agent request: {str(e)}"}, "status_code": 500}
